# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from unittest import TestCase
from rubymarshal.classes import Symbol

__author__ = 'Matthieu Gallet'


class TestSymbol(TestCase):

    def test_symbols(self):
        a = Symbol('test1')
        b = Symbol('test2')
        c = Symbol('test1')
        self.assertNotEqual(a, b)
        self.assertEqual(a, c)
        self.assertEqual(id(a), id(c))



# -*- coding: utf-8 -*-
"""
Sample test module corresponding to the :mod:`rubymarshal.sample` module.

A complete documentation can be found at :mod:`unittest`.

"""

from __future__ import unicode_literals
import re
import unittest
from unittest.case import TestCase
import math

from rubymarshal.classes import Symbol, UsrMarshal
from rubymarshal.reader import loads

__author__ = 'Matthieu Gallet'


class TestLong(TestCase):

    def test_0(self):
        self.assertEqual(loads(b'\x04\bi\x00'), 0)

    def test_1(self):
        self.assertEqual(loads(b'\x04\bi\x06'), 1)

    def test_122(self):
        self.assertEqual(loads(b'\x04\bi\x7F'), 122)

    def test_123(self):
        self.assertEqual(loads(b'\x04\bi\x01{'), 123)

    def test_255(self):
        self.assertEqual(loads(b'\x04\bi\x01\xFF'), 255)

    def test_256(self):
        self.assertEqual(loads(b'\x04\bi\x02\x00\x01'), 256)

    def test_65535(self):
        self.assertEqual(loads(b'\x04\bi\x02\xFF\xFF'), 65535)

    def test_65536(self):
        self.assertEqual(loads(b'\x04\bi\x03\x00\x00\x01'), 65536)

    def test_65537000(self):
        self.assertEqual(loads(b"\x04\bi\x04\xE8\x03\xE8\x03"), 65537000)

    def test__1(self):
        self.assertEqual(loads(b'\x04\bi\xFA'), -1)

    def test__123(self):
        self.assertEqual(loads(b'\x04\bi\x80'), -123)

    def test__124(self):
        self.assertEqual(loads(b'\x04\bi\xFF\x84'), -124)

    def test__256(self):
        self.assertEqual(loads(b'\x04\bi\xFF\x00'), -256)
        self.assertEqual(loads(b'\x04\bi\xFE\x00\xFF'), -256)

    def test__257(self):
        self.assertEqual(loads(b'\x04\bi\xFE\xFF\xFE'), -257)

    def test__259(self):
        self.assertEqual(loads(b"\x04\bi\xFE\xFD\xFE"), -259)

    def test__65536(self):
        self.assertEqual(loads(b'\x04\bi\xFE\x00\x00'), -65536)

    def test__65537(self):
        self.assertEqual(loads(b'\x04\bi\xFD\xFF\xFF\xFE'), -65537)

    def test__65537000(self):
        self.assertEqual(loads(b"\x04\bi\xFC\x18\xFC\x17\xFC"), -65537000)

    def test_long(self):
        self.assertEqual(loads(b"\x04\bl+\t\x15\x81\xE9}\xF4\x10\"\x11"), 1234567890123456789)

    def test_longlong(self):
        self.assertEqual(loads(b"\x04\bl+\x16\xD0\xE8\xDD\x86T\x9D$\b&\xF83E\xE3\xD2\xFD\xB9\"\x12\r\x85\"\v\x92\x06\xEF\x7F}/\xB7_\xB7\xEF\xA5)"),
                         1234567890123456789012343294802948320948209482309842309483209482309482309482309840)


class TestString(TestCase):

    def test_0(self):
        self.assertEqual(loads(b'\x04\bI\"\x00\x06:\x06ET'), '')

    def test_1(self):
        self.assertEqual(loads(b'\x04\bI\"\x06a\x06:\x06ET'), 'a')

    def test_5(self):
        self.assertEqual(loads(b'\x04\bI\"\nabcde\x06:\x06ET'), 'abcde')

    def test_utf32(self):
        self.assertEqual(loads(b"\x04\bI\"\x15\x00\x00\xFE\xFF\x00\x00\x00a\x00\x00\x00b\x00\x00\x00c\x06:\rencoding\"\vUTF-32"),  "abc")
        self.assertEqual(loads(b"\x04\bI\"\r\x00\x00\xFE\xFF\x00\x00'\x13\x06:\rencoding\"\vUTF-32"), '✓')

    def test_utf16(self):
        self.assertEqual(loads(b"\x04\bI\"\r\xFE\xFF\x00a\x00b\x00c\x06:\rencoding\"\vUTF-16"),  "abc")


class TestNil(TestCase):

    def test_nil(self):
        self.assertEqual(loads(b'\x04\b0'), None)


class TestBool(TestCase):

    def test_true(self):
        self.assertEqual(loads(b'\x04\bT'), True)

    def test_false(self):
        self.assertEqual(loads(b'\x04\bF'), False)


class TestArray(TestCase):

    def test_num_3(self):
        self.assertEqual(loads(b"\x04\b[\bi\x06i\ai\b"), [1, 2, 3])

    def test_recursive(self):
        self.assertEqual(loads(b"\x04\b[\ni\x06i\ai\b[\bI\"\ttest\x06:\x06ETF0[\x06i/"), [1, 2, 3, ["test", False, None], [42]])


class TestHash(TestCase):

    def test_num_2(self):
        self.assertEqual(loads(b"\x04\b{\ai\x06i\ai\bi\t"), {1: 2, 3: 4})

    def test_recursive(self):
        self.assertEqual(loads(b"\x04\b{\ni\x06i\ai\bi\ti\n[\bi\x06i\ai\bi\v{\x06i\fi\ri\x0EI\"\ttest\x06:\x06ET"),
                         {1: 2, 3: 4, 5: [1, 2, 3], 6: {7: 8}, 9: "test"})


class TestFloat(TestCase):

    def test_num_0(self):
        self.assertEqual(loads(b"\x04\bf\x060"), 0.0)

    def test_num_1(self):
        self.assertEqual(loads(b"\x04\bf\x061"), 1.0)

    def test_num_1_2(self):
        self.assertEqual(loads(b"\x04\bf\b1.2"), 1.2)

    def test_num__1(self):
        self.assertEqual(loads(b"\x04\bf\a-1"), -1.0)

    def test_num_1234567890_123456789(self):
        self.assertEqual(loads(b"\x04\bf\x171234567890.1234567"), 1234567890.1234567)

    def test_num_inf(self):
        self.assertEqual(loads(b"\x04\bf\binf"), float('inf'))

    def test_num_nan(self):

        self.assertTrue(math.isnan(loads(b"\x04\bf\bnan")))

    def test_num__inf(self):
        self.assertEqual(loads(b"\x04\bf\t-inf"), float('-inf'))


class TestRegexp(TestCase):
    def test_noflag(self):
        self.assertEqual(loads(b"\x04\bI/\att\x00\x06:\x06EF"), re.compile('tt'))

    def test_flag_1(self):
        self.assertEqual(loads(b"\x04\bI/\att\x01\x06:\x06EF"), re.compile('tt', re.IGNORECASE))
        self.assertNotEqual(loads(b"\x04\bI/\att\x01\x06:\x06EF"), re.compile('tt'))

    def test_flag_4(self):
        self.assertEqual(loads(b"\x04\bI/\att\x04\x06:\x06EF"), re.compile('tt', re.DOTALL))
        self.assertNotEqual(loads(b"\x04\bI/\att\x04\x06:\x06EF"), re.compile('tt'))


class TestUsrMarshal(TestCase):

    def test_usr(self):
        a = UsrMarshal('Gem::Version', ['0.1.2'])
        self.assertEqual(loads(b"\x04\bU:\x11Gem::Version[\x06I\"\n0.1.2\x06:\x06ET"), a)


class TestSymbol(TestCase):
    def test_symbol(self):
        self.assertEqual(loads(b"\x04\b:\x10test_symbol"), Symbol('test_symbol'))


class TestSymlink(TestCase):
    def test_symlink(self):
        a = Symbol('test_symbol')
        self.assertEqual(loads(b"\x04\b[\a:\x10test_symbol;\x00"), [a, a])


class TestLink(TestCase):

    def test_link_base(self):
        a = [1, 2, 3]
        result = loads(b"\x04\b[\b[\bi\x06i\ai\b@\x06@\x06")
        self.assertEqual(result, [a, a, a])
        result[0][2] = 4
        self.assertEqual(result[1][2], 4)
        self.assertEqual(result[2][2], 4)

    def test_link_usr(self):
        a = UsrMarshal('Gem::Version', ['0.1.2'])
        result = loads(b"\x04\b[\aU:\x11Gem::Version[\x06I\"\n0.1.2\x06:\x06ET@\x06")
        self.assertEqual(result, [a, a])

    def test_link_usr_base(self):
        a = [1, 2, 3]
        b = UsrMarshal('Gem::Version', ['0.1.2'])
        result = loads(b"\x04\b[\b[\a[\bi\x06i\ai\b@\a[\aU:\x11Gem::Version[\x06I\"\n0.1.2\x06:\x06ET@\t[\a@\a@\a")
        self.assertEqual(result, [[a, a], [b, b], [a, a]])

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import division, unicode_literals
import re
import io
import math

from rubymarshal.classes import Symbol, UsrMarshal
from rubymarshal.constants import TYPE_BIGNUM, TYPE_STRING, TYPE_REGEXP, TYPE_ARRAY, TYPE_HASH, TYPE_USRMARSHAL, TYPE_NIL, TYPE_TRUE, TYPE_FALSE, \
    TYPE_IVAR, TYPE_LINK, TYPE_SYMLINK, TYPE_SYMBOL, TYPE_FIXNUM
from rubymarshal.constants import TYPE_FLOAT
from rubymarshal.utils import write_ushort, write_sbyte, write_ubyte, integer_types, binary_type, text_type

__author__ = 'Matthieu Gallet'

re_class = re.compile('').__class__
simple_float_re = re.compile(r'^\d+\.\d*0+$')


class Writer(object):
    def __init__(self, fd):
        self.symbols = {}
        self.objects = {}
        self.fd = fd

    def write(self, obj):

        if obj is None:
            self.fd.write(TYPE_NIL)
        elif obj is False:
            self.fd.write(TYPE_FALSE)
        elif obj is True:
            self.fd.write(TYPE_TRUE)
        elif isinstance(obj, int) or isinstance(obj, integer_types[-1]):
            if obj.bit_length() <= 5 * 8:
                self.fd.write(TYPE_FIXNUM)
                # noinspection PyTypeChecker
                self.write_long(obj)
            else:
                self.fd.write(TYPE_BIGNUM)
                if obj < 0:
                    self.fd.write(b'-')
                else:
                    self.fd.write(b'+')
                obj = abs(obj)
                size = int(math.ceil(obj.bit_length() / 16.))
                self.write_long(size)
                for i in range(size):
                    self.write_short(obj % 65536)
                    obj //= 65536
        elif isinstance(obj, Symbol):
            if obj.name in self.symbols:
                self.fd.write(TYPE_SYMLINK)
                self.write_long(self.symbols[obj.name])
            else:
                self.fd.write(TYPE_SYMBOL)
                symbol_index = len(self.symbols)
                self.symbols[obj.name] = symbol_index
                encoded = obj.name.encode('utf-8')
                self.write_long(len(encoded))
                self.fd.write(encoded)
        elif isinstance(obj, list):
            if self.must_write(obj):
                self.fd.write(TYPE_ARRAY)
                self.write_long(len(obj))
                for x in obj:
                    self.write(x)
        elif isinstance(obj, dict):
            if self.must_write(obj):
                self.fd.write(TYPE_HASH)
                self.write_long(len(obj))
                for key, value in obj.items():
                    self.write(key)
                    self.write(value)
        elif isinstance(obj, binary_type):
            self.fd.write(TYPE_IVAR)
            self.fd.write(TYPE_STRING)
            self.write_long(len(obj))
            self.fd.write(obj)
            self.write_long(1)
            self.write(Symbol('E'))
            self.write(False)
        elif isinstance(obj, text_type):
            obj = obj.encode('utf-8')
            self.fd.write(TYPE_IVAR)
            self.fd.write(TYPE_STRING)
            self.write_long(len(obj))
            self.fd.write(obj)
            self.write_long(1)
            self.write(Symbol('E'))
            self.write(True)
        elif isinstance(obj, float):
            obj = '%.20g' % obj
            if simple_float_re.match(obj):
                while obj.endswith('0'):
                    obj = obj[:-1]
            obj = obj.encode('utf-8')
            self.fd.write(TYPE_FLOAT)
            self.write_long(len(obj))
            self.fd.write(obj)
        elif isinstance(obj, re_class):
            flags = 0
            if obj.flags & re.IGNORECASE:
                flags += 1
            if obj.flags & re.DOTALL:
                flags += 4
            self.fd.write(TYPE_IVAR)
            self.fd.write(TYPE_REGEXP)
            pattern = obj.pattern.encode('utf-8')
            self.write_long(len(pattern))
            self.fd.write(pattern)
            write_ubyte(self.fd, flags)
            self.write_long(1)
            self.write(Symbol('E'))
            self.write(False)
        elif isinstance(obj, UsrMarshal):
            if self.must_write(obj):
                self.fd.write(TYPE_USRMARSHAL)
                self.write(Symbol(obj.cls))
                self.write(obj.values)
        else:
            raise ValueError(obj)

    def write_short(self, obj):
        write_ushort(self.fd, obj)

    def write_long(self, obj):
        if obj == 0:
            self.fd.write(b'\0')
        elif 0 < obj < 123:
            write_sbyte(self.fd, obj + 5)
        elif -124 < obj < 0:
            write_sbyte(self.fd, obj - 5)
        else:
            size = int(math.ceil(obj.bit_length() / 8.))
            if size > 5:
                raise ValueError('%d too long for serialization' % obj)
            original_obj = obj
            factor = 256 ** size
            if obj < 0 and obj == -factor:
                size -= 1
                obj += (factor / 256)
            elif obj < 0:
                obj += factor
            sign = int(math.copysign(size, original_obj))
            write_sbyte(self.fd, sign)
            for i in range(size):
                write_ubyte(self.fd, obj % 256)
                obj //= 256

    def must_write(self, obj):
        if id(obj) in self.objects:
            self.fd.write(TYPE_LINK)
            self.write_long(self.objects[id(obj)])
            return False
        else:
            link_index = len(self.objects)
            self.objects[id(obj)] = link_index
            return True


def write(fd, obj):
    fd.write(b'\x04\x08')
    writer = Writer(fd)
    writer.write(obj)


def writes(obj):
    fd = io.BytesIO()
    write(fd, obj)
    return fd.getvalue()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import re
from unittest import TestCase
import io
import math
from rubymarshal.classes import UsrMarshal, Symbol
from rubymarshal.reader import loads
from rubymarshal.writer import Writer, writes

__author__ = 'Matthieu Gallet'


def long_write(obj):
    fd = io.BytesIO()
    writer = Writer(fd)
    writer.write_long(obj)
    return fd.getvalue()


class TestWriteLong(TestCase):
    def test_0(self):
        self.assertEqual(b'\x00', long_write(0))

    def test_1(self):
        self.assertEqual(b'\x06', long_write(1))

    def test_122(self):
        self.assertEqual(b'\x7F', long_write(122))

    def test_123(self):
        self.assertEqual(b'\x01{', long_write(123))

    def test_255(self):
        self.assertEqual(b'\x01\xFF', long_write(255))

    def test_256(self):
        self.assertEqual(b'\x02\x00\x01', long_write(256))

    def test_65535(self):
        self.assertEqual(b'\x02\xFF\xFF', long_write(65535))

    def test_65536(self):
        self.assertEqual(b'\x03\x00\x00\x01', long_write(65536))

    def test_65537000(self):
        self.assertEqual(b"\x04\xE8\x03\xE8\x03", long_write(65537000))

    def test__1(self):
        self.assertEqual(b'\xFA', long_write(-1))

    def test__123(self):
        self.assertEqual(b'\x80', long_write(-123))

    def test__124(self):
        self.assertEqual(b'\xFF\x84', long_write(-124))

    def test__256(self):
        self.assertEqual(b'\xfe\x00\xff', long_write(-256))

    def test__512(self):
        self.assertEqual(b'\xfe\x00\xfe', long_write(-512))

    def test__768(self):
        self.assertEqual(b"\xFE\x00\xFD", long_write(-768))

    def test__257(self):
        self.assertEqual(b'\xFE\xFF\xFE', long_write(-257))

    def test__259(self):
        self.assertEqual(b"\xFE\xFD\xFE", long_write(-259))

    def test__65536(self):
        self.assertEqual(b'\xfd\x00\x00\xff', long_write(-65536))

    def test__65537(self):
        self.assertEqual(b'\xFD\xFF\xFF\xFE', long_write(-65537))

    def test__65537000(self):
        self.assertEqual(b"\xFC\x18\xFC\x17\xFC", long_write(-65537000))


class TestIdemPotent(TestCase):
    def read_write(self, x):
        self.assertEqual(loads(writes(x)), x)


class TestLong(TestIdemPotent):

    def test_0(self):
        self.read_write(0)

    def test_1(self):
        self.read_write(1)

    def test_122(self):
        self.read_write(122)

    def test_123(self):
        self.read_write(123)

    def test_255(self):
        self.read_write(255)

    def test_256(self):
        self.read_write(256)

    def test_65536(self):
        self.read_write(65536)

    def test_65535(self):
        self.read_write(65535)

    def test_65537000(self):
        self.read_write(65537000)

    def test__1(self):
        self.read_write(-1)

    def test__123(self):
        self.read_write(-123)

    def test__124(self):
        self.read_write(-124)

    def test__256(self):
        self.read_write(-256)

    def test__257(self):
        self.read_write(-257)

    def test__259(self):
        self.read_write(-259)

    def test__65536(self):
        self.read_write(-65536)

    def test__65537(self):
        self.read_write(-65537)

    def test__65537000(self):
        self.read_write(-65537000)

    def test_long(self):
        self.read_write(1234567890123456789)

    def test_longlong(self):
        self.read_write(1234567890123456789012343294802948320948209482309842309483209482309482309482309840)


class TestNil(TestIdemPotent):

    def test_nil(self):
        self.read_write(None)


class TestBool(TestIdemPotent):

    def test_true(self):
        self.read_write(True)

    def test_false(self):
        self.read_write(False)


class TestArray(TestIdemPotent):

    def test_num_3(self):
        self.read_write([1, 2, 3])

    def test_recursive(self):
        self.read_write([1, 2, 3, [False, None], [42]])


class TestString(TestIdemPotent):

    def test_0(self):
        self.read_write('')

    def test_1(self):
        self.read_write('a')

    def test_5(self):
        self.read_write('abcde')

    def test_unicode(self):
        self.read_write('✓')


class TestHash(TestIdemPotent):

    def test_num_2(self):
        self.read_write({1: 2, 3: 4})

    def test_recursive(self):
        self.read_write({1: 2, 3: 4, 5: [1, 2, 3], 6: {7: 8}, 9: "test"})


class TestFloat(TestIdemPotent):

    def test_num_0(self):
        self.read_write(0.0)

    def test_num_1(self):
        self.read_write(1.0)

    def test_num_1_2(self):
        self.read_write(1.2)

    def test_num__1(self):
        self.read_write(-1.0)

    def test_num_1234567890_123456789(self):
        self.read_write(1234567890.1234567)

    def test_num_inf(self):
        self.read_write(float('inf'))

    def test_num_nan(self):
        self.assertTrue(math.isnan(loads(writes(float('nan')))))

    def test_num__inf(self):
        self.read_write(float('-inf'))


class TestRegexp(TestIdemPotent):
    def test_noflag(self):

        self.assertEqual(b"\x04\bI/\att\x00\x06:\x06EF", writes(re.compile('tt')))
        self.assertEqual(b"\x04\bI/\att\x01\x06:\x06EF", writes(re.compile('tt', re.IGNORECASE)))

        self.read_write(re.compile('tt'))

    def test_flag(self):
        self.read_write(re.compile('tt', re.IGNORECASE))
        self.read_write(re.compile('tt', re.DOTALL))


class TestUsrMarshal(TestIdemPotent):

    def test_usr(self):
        a = UsrMarshal('Gem::Version', ['0.1.2'])
        self.read_write(a)


class TestSymbol(TestIdemPotent):
    def test_symbol(self):
        self.read_write(Symbol('test_symbol'))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
__author__ = 'Matthieu Gallet'


class RubyClass(object):
    def __init__(self, cls, values):
        if isinstance(cls, Symbol):
            cls = cls.name
        self.cls = cls
        self.values = values

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.cls == other.cls and self.values == other.values

    def __hash__(self):
        return hash(hash(self.cls) + hash(self.values))

    def __repr__(self):
        return '%s:%s(%r)' % (self.__class__.__name__, self.cls, self.values)

    def __str__(self):
        return '%s:%s(%r)' % (self.__class__.__name__, self.cls, self.values)

    def __unicode__(self):
        return '%s:%s(%r)' % (self.__class__.__name__, self.cls, self.values)


class UsrMarshal(RubyClass):
    pass


class UserDef(RubyClass):
    pass


class Extended(RubyClass):
    pass


class Module(RubyClass):
    pass


class Object(RubyClass):
    def __init__(self, cls, values, **kwargs):
        self.kwargs = kwargs
        super(Object, self).__init__(cls, values)


class Symbol(object):

    __registered_symbols__ = {}

    def __new__(cls, name):
        if name in cls.__registered_symbols__:
            return cls.__registered_symbols__[name]
        return super(Symbol, cls).__new__(cls)

    def __init__(self, name):
        self.name = name
        self.__registered_symbols__[name] = self

    def __hash__(self):
        return hash('<<<:%s:>>>' % self.name)

    def __repr__(self):
        return 'Symbol("%s")' % self.name

    def __str__(self):
        return ':%s' % self.name

    def __unicode__(self):
        return ':%s' % self.name



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from unittest import TestCase
import io
from rubymarshal.utils import write_ubyte, write_sbyte, write_ushort, read_ushort, read_sbyte, read_ubyte

__author__ = 'Matthieu Gallet'


class Test(TestCase):

    def check_write(self, fn, obj):
        fd = io.BytesIO()
        fn(fd, obj)
        return fd.getvalue()

    def check_read(self, fn, content):
        fd = io.BytesIO(content)
        return fn(fd)

    def test_write_ushort(self):
        content = self.check_write(write_ushort, 1)
        self.assertEqual(b'\x01\x00', content)

    def test_write_sbyte(self):
        content = self.check_write(write_sbyte, 1)
        self.assertEqual(b'\x01', content)
        content = self.check_write(write_sbyte, -1)
        self.assertEqual(b'\xff', content)

    def test_write_ubyte(self):
        content = self.check_write(write_ubyte, 1)
        self.assertEqual(b'\x01', content)
        content = self.check_write(write_ubyte, 255)
        self.assertEqual(b'\xff', content)

    def test_read_ushort(self):
        content = self.check_read(read_ushort, b'\x01\x00')
        self.assertEqual(1, content)

    def test_read_sbyte(self):
        content = self.check_read(read_sbyte, b'\x01')
        self.assertEqual(1, content)
        content = self.check_read(read_sbyte, b'\xff')
        self.assertEqual(-1, content)

    def test_read_ubyte(self):
        content = self.check_read(read_ubyte, b'\x01')
        self.assertEqual(1, content)
        content = self.check_read(read_ubyte, b'\xff')
        self.assertEqual(255, content)
#


# -*- coding: utf-8 -*-
"""regroup all struct functions and some compatibility tricks for Python2/3.

Do not import unicode_literals from __future__.
"""
import sys
import struct

__author__ = 'Matthieu Gallet'

# Useful for very coarse version differentiation.
PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY3:
    integer_types = int,
    text_type = str
    binary_type = bytes
else:
    # noinspection PyUnresolvedReferences
    integer_types = (int, long)
    # noinspection PyUnresolvedReferences
    text_type = unicode
    binary_type = str


def write_ushort(fd, obj):
    fd.write(struct.pack('<H', obj))


def write_sbyte(fd, obj):
    fd.write(struct.pack('b', obj))


def write_ubyte(fd, obj):
    fd.write(struct.pack('B', obj))


def read_ushort(fd):
    return struct.unpack('<H', fd.read(2))[0]


def read_sbyte(fd):
    return struct.unpack('b', fd.read(1))[0]


def read_ubyte(fd):
    return struct.unpack('B', fd.read(1))[0]



# -*- coding: utf-8 -*-
from unittest import TestCase
# noinspection PyPackageRequirements
from hypothesis import given
# noinspection PyPackageRequirements
from hypothesis.strategies import integers, lists, booleans, floats, dictionaries, text
import math
from rubymarshal.reader import loads
from rubymarshal.writer import writes

__author__ = 'Matthieu Gallet'


class TestValues(TestCase):
    def read_write(self, x):
        self.assertEqual(loads(writes(x)), x)

    @given(integers())
    def test_ints(self, x):
        self.read_write(x)

    @given(lists(integers()))
    def test_lists(self, x):
        self.read_write(x)

    @given(lists(booleans()))
    def test_booleans(self, x):
        self.read_write(x)

    @given(floats())
    def test_floats(self, x):
        if math.isnan(x):
            self.assertTrue(math.isnan(loads(writes(x))))
        else:
            self.read_write(x)

    @given(dictionaries(integers(), integers()))
    def test_dictionaries(self, x):
        self.read_write(x)

    @given(lists(text()))
    def test_text(self, x):
        self.read_write(x)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
__author__ = 'Matthieu Gallet'

TYPE_NIL = b'0'
TYPE_TRUE = b'T'
TYPE_FALSE = b'F'
TYPE_FIXNUM = b'i'
TYPE_ARRAY = b'['
TYPE_SYMBOL = b':'
TYPE_STRING = b'"'
TYPE_IVAR = b'I'
TYPE_HASH = b'{'
TYPE_FLOAT = b'f'
TYPE_BIGNUM = b'l'
TYPE_REGEXP = b'/'
TYPE_USRMARSHAL = b'U'
TYPE_SYMLINK = b';'
TYPE_LINK = b'@'


TYPE_EXTENDED = b'e'
TYPE_UCLASS = b'C'
TYPE_OBJECT = b'o'
TYPE_DATA = b'd'
TYPE_USERDEF = b'u'
TYPE_HASH_DEF = b'}'
TYPE_STRUCT = b'S'
TYPE_MODULE_OLD = b'M'
TYPE_CLASS = b'c'
TYPE_MODULE = b'm'



# -*- coding: utf-8 -*-
"""
Package gathering all unitary tests for rubymarshal.
Module names must start with `test_` to be taken into account.

You should consider to install :mod:`Distribute` to run all tests with::

    $ python setup.py test

"""
from __future__ import unicode_literals
__author__ = 'Matthieu Gallet'



# -*- coding: utf-8 -*-
"""Setup file for the RubyMarshal project.
"""

import codecs
import os.path
import re

from setuptools import setup, find_packages


with codecs.open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf-8') as fd:
    long_description = fd.read()
version = None
for line in codecs.open(os.path.join('rubymarshal', '__init__.py'), 'r', encoding='utf-8'):
    matcher = re.match(r"""^__version__\s*=\s*['"](.*)['"]\s*$""", line)
    version = version or matcher and matcher.group(1)

setup(
    name='rubymarshal',
    version=version,
    description='Read and write Ruby-marshalled data.',
    long_description=long_description,
    author='Matthieu Gallet',
    author_email='python-dev@19pouces.net',
    license='WTFPL',
    url='https://github.com/d9pouces/RubyMarshal',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=True,
    test_suite='rubymarshal.tests',
    install_requires=[],
    setup_requires=[],
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Intended Audience :: System Administrators',
                 'License :: OSI Approved',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.2',
                 'Programming Language :: Python :: 3.3',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Topic :: Utilities',
                 ],
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import re
import io
from rubymarshal.classes import UsrMarshal, Symbol, UserDef, Extended, Object, Module

from rubymarshal.constants import TYPE_NIL, TYPE_TRUE, TYPE_FALSE, TYPE_FIXNUM, TYPE_IVAR, TYPE_STRING, TYPE_SYMBOL, TYPE_ARRAY, TYPE_HASH, \
    TYPE_FLOAT, TYPE_BIGNUM, TYPE_REGEXP, TYPE_USRMARSHAL, \
    TYPE_SYMLINK, TYPE_LINK, TYPE_DATA, TYPE_OBJECT, TYPE_STRUCT, TYPE_MODULE, TYPE_CLASS, TYPE_USERDEF, TYPE_EXTENDED
from rubymarshal.utils import read_ushort, read_sbyte, read_ubyte

__author__ = 'Matthieu Gallet'


class Reader(object):
    def __init__(self, fd, usr_marshal_mapping=None, userdef_mapping=None):
        self.symbols = []
        self.objects = []
        self.fd = fd
        self.usr_marshal_mapping = usr_marshal_mapping or {}
        self.userdef_mapping = userdef_mapping or {}

    def read(self, token=None, ivar=False):
        if token is None:
            token = self.fd.read(1)
        object_index = None
        if token in (TYPE_CLASS, TYPE_MODULE, TYPE_FLOAT, TYPE_BIGNUM, TYPE_STRING, TYPE_REGEXP, TYPE_ARRAY, TYPE_HASH, TYPE_STRUCT, TYPE_OBJECT, TYPE_DATA, TYPE_USRMARSHAL):
            self.objects.append(None)
            object_index = len(self.objects)
        if token == TYPE_NIL:
            result = None
        elif token == TYPE_TRUE:
            result = True
        elif token == TYPE_FALSE:
            result = False
        elif token == TYPE_IVAR:
            sub_token = self.fd.read(1)
            content = self.read(sub_token, ivar=True)
            flags = None
            if sub_token != TYPE_STRING:
                print(token,  sub_token, '-----')
            if sub_token == TYPE_REGEXP:
                options = ord(self.fd.read(1))
                flags = 0
                if options & 1:
                    flags |= re.IGNORECASE
                if options & 4:
                    flags |= re.DOTALL
            if sub_token in (TYPE_STRING, TYPE_REGEXP, TYPE_USERDEF):
                encoding = 'latin1'
                attr_count = self.read_long()
                attrs = {}
                for x in range(attr_count):
                    attr_name = self.read()
                    attr_value = self.read()
                    if attr_name == Symbol('E') and attr_value is True:
                        encoding = 'utf-8'
                    elif attr_name == Symbol('encoding'):
                        encoding = attr_value.decode('utf-8')
                    attrs[attr_name] = attr_value
                if sub_token in (TYPE_STRING, TYPE_REGEXP):
                    content = content.decode(encoding)
            else:
                raise ValueError
            if sub_token == TYPE_REGEXP:
                content = re.compile(content, flags)
            result = content
        elif token == TYPE_STRING:
            size = self.read_long()
            result = self.fd.read(size)
        elif token == TYPE_SYMBOL:
            result = self.read_symreal()
        elif token == TYPE_FIXNUM:
            result = self.read_long()
        elif token == TYPE_ARRAY:
            num_elements = self.read_long()
            # noinspection PyUnusedLocal
            result = [self.read() for x in range(num_elements)]
        elif token == TYPE_HASH:
            num_elements = self.read_long()
            result = {}
            for x in range(num_elements):
                key = self.read()
                value = self.read()
                result[key] = value
            result = result
        elif token == TYPE_FLOAT:
            size = self.read_long()
            result = float(self.fd.read(size).decode('utf-8'))
        elif token == TYPE_BIGNUM:
            sign = 1 if self.fd.read(1) == b'+' else -1
            num_elements = self.read_long()
            result = 0
            factor = 1
            for x in range(num_elements):
                result += (self.read_short() * factor)
                factor *= 2 ** 16
            result *= sign
        elif token == TYPE_REGEXP:
            size = self.read_long()
            result = self.fd.read(size)
        elif token == TYPE_USRMARSHAL:
            class_name = self.read()
            attr_list = self.read()
            cls = self.usr_marshal_mapping.get(class_name, UsrMarshal)
            result = cls(class_name, attr_list)
        elif token == TYPE_SYMLINK:
            result = self.read_symlink()
        elif token == TYPE_LINK:
            link_id = self.read_long()
            result = self.objects[link_id]
        elif token == TYPE_USERDEF:
            class_name = self.read()
            data = self.read(TYPE_STRING)
            if not ivar:
                data = loads(data)
            result = UserDef(class_name, data)
            print(result)
        elif token == TYPE_MODULE:
            module_name = self.read()
            result = Module(module_name, None)
        elif token == TYPE_OBJECT:
            class_name = self.read()
            data = self.read(TYPE_STRING)
            # attr_count = self.read_long()
            # # noinspection PyUnusedLocal
            # attrs = {self.read(): self.read() for x in range(attr_count)}
            if not ivar:
                data = self.read()
            result = Object(class_name, data)
            print(result)
        elif token == TYPE_EXTENDED:
            class_name = self.read(TYPE_STRING)
            print(class_name)
            result = Extended(class_name, None)
        else:
            raise ValueError('token %s is not recognized' % token)
        if object_index is not None:
            self.objects[object_index - 1] = result
        # print('end', result, object_index, self.objects)
        return result

    def read_short(self):
        return read_ushort(self.fd)

    def read_long(self):
        l = read_sbyte(self.fd)
        if l == 0:
            return 0
        if 5 < l < 128:
            return l - 5
        elif -129 < l < -5:
            return l + 5
        result = 0
        factor = 1
        for s in range(abs(l)):
            result += (read_ubyte(self.fd) * factor)
            factor *= 256
        if l < 0:
            result = result - factor
        return result

    def read_symbol(self):
        ivar = 0
        while True:
            token = self.fd.read(1)
            if token == TYPE_IVAR:
                ivar = 1
            elif token == TYPE_SYMBOL:
                return self.read_symreal()
            elif token == TYPE_SYMLINK:
                if ivar:
                    raise ValueError('dump format error (symlink with encoding)')
                return self.read_symlink()

    def read_symlink(self):
        symlink_id = self.read_long()
        return self.symbols[symlink_id]

    def read_symreal(self):
        size = self.read_long()
        result = self.fd.read(size)
        result = Symbol(result.decode('utf-8'))
        self.symbols.append(result)
        return result


def load(fd):
    assert fd.read(1) == b'\x04'
    assert fd.read(1) == b'\x08'
    loader = Reader(fd)
    return loader.read()


def loads(byte_text):
    return load(io.BytesIO(byte_text))
